// ------------------------------

class Either<A, B> {}

class Left<A, B> implements Either<A, B> {
  Left(this.value);
  A value;
}

class Right<A, B> implements Either<A, B> {
  Right(this.value);
  B value;
}

abstract class Option<A> {
  const Option();
  B Function(B Function(A)) fold<B>(B ifEmpty()) =>
      (B f(A a)) => isEmpty ? ifEmpty() : f(value);
  bool get isEmpty;
  A get value;
}

class Some<A> extends Option<A> {
  Some(this.value);
  @override
  bool get isEmpty => false;
  @override
  final A value;
}

class _None extends Option<Null> {
  //Never
  const _None();

  @override
  bool get isEmpty {
    return true;
  }

  @override
  Null get value {
    throw "Attempt to do None.value";
  }
}

const None = _None();

// ------------------------------ TailCall.

typedef LazyTailRec<X> = TailRec<X> Function();

abstract class TailRec<A> {
  TailRec<B> map<B>(B Function(A) f) =>
      flatMap((a) => _Call(() => _Done(f(a))));

  // abstract
  TailRec<B> flatMap<B>(TailRec<B> Function(A) f);

  // abstract
  A get result;

  // abstract
  Either<LazyTailRec<A>, A> get _resume;
  
  // abstract
  B _result1<B>(_Cont<A, B> c);
  
  // abstract
  Either<LazyTailRec<B>, B> _resume1<B>(_Cont<A, B> c);

  static C _result0<D, C>(TailRec<D> b, _Cont<D, C> c) => b._result1(c);

  
  static Either<LazyTailRec<C>, C> _resume0<D, C>(
          TailRec<D> b, _Cont<D, C> c) =>
      b._resume1(c);
}

class _Call<A> extends TailRec<A> {
  _Call(this.rest);
  final LazyTailRec<A> rest;

  @override
  TailRec<B> flatMap<B>(f) => _Cont(this, f);
  @override
  A get result => rest().result;
  @override
  Left<TailRec<A> Function(), A> get _resume => Left(rest);
  @override
  B _result1<B>(c) => rest().flatMap(c.f).result;
  @override
  Left<TailRec<B> Function(), B> _resume1<B>(c) =>
      Left(() => rest().flatMap(c.f));
}

class _Cont<B, A> extends TailRec<A> {
  _Cont(this.b, this.f);
  TailRec<B> b;
  TailRec<A> Function(B) f;

  @override
  TailRec<C> flatMap<C>(g) => _Cont(b, (B x) => f(x).flatMap(g));

  @override
  A get result => TailRec._result0<B, A>(b, this);

  @override
  Either<TailRec<A> Function(), A> get _resume =>
      TailRec._resume0<B, A>(b, this);
  @override
  C _result1<C>(c) => b.flatMap((B x) => f(x).flatMap(c.f)).result;
  @override
  Either<TailRec<C> Function(), C> _resume1<C>(c) =>
      b.flatMap((B x) => f(x).flatMap(c.f))._resume;
}

class _Done<A> extends TailRec<A> {
  _Done(this.value);
  final A value;

  @override
  _Call<B> flatMap<B>(f) => _Call(() => f(value));
  @override
  A get result => value;
  @override
  Right<TailRec<A> Function(), A> get _resume => Right(value);
  @override
  B _result1<B>(c) => c.f(value).result;
  @override
  Either<TailRec<B> Function(), B> _resume1<B>(c) => c.f(value)._resume;
}
/*

Original

TailRec<A> tailcall<A>(TailRec<A> rest) => _Call(() => rest);
TailRec<A> done<A>(A result) => _Done(result);

*/
/* ------------------------------------------------ */

TailRec<A> tailcall<A>(TailRec<A> Function() rest) => _Call(rest);
TailRec<A> done<A>(A result) => _Done(result);

// kein Fibonacci
TailRec<int> rec(int desiredCount) {
  TailRec<int> itar(int a, int b, int count) {
    if (count == desiredCount) {
      return done(a + b);
    } else {
      return tailcall(() => itar(b, a + b, count + 1));
    }
  }

  return itar(0, 1, 0); // recursive call
}

TailRec<int> fib(int n) {
  if (n < 2) {
    return done(n);
  } else {
    return tailcall(() => fib(n - 1)).flatMap((x) {
      return tailcall(() => fib(n - 2)).map((y) {
        return (x + y);
      });
    });
  }
}

void main() {
  print("Recursion ${rec(10).result};");
  for (int i = 0; i < 20; i++) {
    print("Fibonacci für ${i} ist ${fib(i).result};");
  }
}
